/*
   ______               _                  _///  _           _                   _
  /   _  \             (_)                |  __\| |         | |                 (_)
  |  [_|  |__  ___  ___ _  ___  _ __      | |__ | | ___  ___| |_ _ __ ___  _ __  _  ___  _   _  ___
  |   ___/ _ \| __|| __| |/ _ \| '_ \_____|  __|| |/ _ \/  _|  _| '__/   \| '_ \| |/   \| | | |/ _ \
  |  |  | ( ) |__ ||__ | | ( ) | | | |____| |__ | |  __/| (_| |_| | | (_) | | | | | (_) | |_| |  __/
  \__|   \__,_|___||___|_|\___/|_| [_|    \____/|_|\___|\____\__\_|  \___/|_| |_|_|\__  |\__,_|\___|
                                                                                      | |
                                                                                      \_|
  Fichier:      prgVoiture3rouesRC
  
  Description:  Programme de bord de l'Arduino Nano, faisant tourner le projet de "Voiture Radiocommandée à 3 roues"
                (exemple de mise en oeuvre des modules émetteur/récepteur NRF24L01+)

                
  Auteur:       Jérôme, Passion-Électronique (https://passionelectronique.fr/)
  Création :    29.11.2021

  Librairie utilisée : https://github.com/nRF24/RF24
  
*/

//***************//
// Bibliothèques //
//***************//
#include <SPI.h>
#include <RF24.h>


//**************************************************//
// Définition des entrées/sorties de l'Arduino Nano //
//**************************************************//
#define borneENA_L298N    9       // On associe la borne "ENA" du L298N à la pin D9 de l'arduino
#define borneIN1_L298N    8       // On associe la borne "IN1" du L298N à la pin D8 de l'arduino
#define borneIN2_L298N    7       // On associe la borne "IN2" du L298N à la pin D7 de l'arduino
#define borneIN3_L298N    5       // On associe la borne "IN3" du L298N à la pin D5 de l'arduino
#define borneIN4_L298N    4       // On associe la borne "IN4" du L298N à la pin D4 de l'arduino
#define borneENB_L298N    3       // On associe la borne "ENB" du L298N à la pin D3 de l'arduino

#define pinCE_RF24       10       // On associe la broche "CE" du NRF24L01 à la pin D10 de l'arduino
#define pinCSN_RF24       2       // On associe la broche "CSN" du NRF24L01 à la pin D2 de l'arduino

#define pinVA_PACK_ACCUS_18650    A0    // [Entrée] Pour mesure de tension "nominale 11,1V" du pack d'accus li-ion 18650
#define pinVB_PACK_ACCUS_18650    A1    // [Entrée] Pour mesure de tension "nominale  7,4V" du pack d'accus li-ion 18650
#define pinVC_PACK_ACCUS_18650    A2    // [Entrée] Pour mesure de tension "nominale  3,7V" du pack d'accus li-ion 18650

#define tensionMinimaleAccu       3     // On définit que la tension minimale de chaque accu 18650 qui alimente se projet, ne doit pas être en dessous de 3V

#define pinLed_BatterieFaible     A3    // [Sortie] Raccordé à la LED rouge présente sur le PCB (batterie faible), s'allumant quand la sortie est mise à 0
#define pinLed_SignalRadioPerdu   A4    // [Sortie] Raccordé à la LED orange présente sur le PCB (signal radio perdu), s'allumant quand la sortie est mise à 0


//*****************************************************//
// Définition de paramètres généraux pour ce programme //
//*****************************************************//
#define tunnelDeCommunication             "REA01"   // Déclaration d'un "nom de tunnel" (5 caractères) => doit être strictement identique à l'émetteur
#define valeurResistanceDiviseur_serie      22000   // Résistance 1/2 du pont diviseur de tension créé pour mesurer la tension de chaque accu 18650 (côté pack accus)
#define valeurResistanceDiviseur_pullDown   10000   // Résistance 2/2 du pont diviseur de tension créé pour mesurer la tension de chaque accu 18650 (côté de ce montage)
#define tensionReferenceAdcArduino              5   // Tension de référence de l'ADC de l'Arduino Nano (par défaut à 5 volts)
#define vitesseMinimale                        55   // Rapport cylique minimal des signaux PWM, pour faire tourner les moteur au minimum de leur vitesse (en pratique, on évitera de trop approcher de la valeur 0)
#define vitesseMaximale                       255   // Rapport cylique maximal des signaux PWM, pour faire tourner les moteur au maximum de leur vitesse


//************************//
// Variables du programme //
//************************//
const byte adresse[6] = tunnelDeCommunication;      // Mise au format "byte array" du nom du tunnel

int valBrocheVA;            // Variable 16 bits où sera stockée la lecture 10 bits (0..1023) de l'ADC en VA
int valBrocheVB;            // Variable 16 bits où sera stockée la lecture 10 bits (0..1023) de l'ADC en VB
int valBrocheVC;            // Variable 16 bits où sera stockée la lecture 10 bits (0..1023) de l'ADC en VC

float valTensionVA;         // Variable de calcul qui donnera la "vraie" valeur de tension de VA, après correction inverse du pont diviseur résistif (11,1V au nominal)
float valTensionVB;         // Variable de calcul qui donnera la "vraie" valeur de tension de VB, après correction inverse du pont diviseur résistif ( 7,4V au nominal)
float valTensionVC;         // Variable de calcul qui donnera la "vraie" valeur de tension de VC, après correction inverse du pont diviseur résistif ( 3,7V au nominal)

float tensionAccu1;         // Variable de calcul qui contiendra la tension de l'accu li-ion n°1
float tensionAccu2;         // Variable de calcul qui contiendra la tension de l'accu li-ion n°2
float tensionAccu3;         // Variable de calcul qui contiendra la tension de l'accu li-ion n°3

unsigned long heureDernierSignalRecu = 0;   // Variable qui contiendra "l'heure" du dernier signal reçu, en provenance de l'émetteur (nota : ce n'est pas réellement l'heure, mais un "compteur de millisecondes écoulées", interne à l'arduino)
boolean bSignalRadioPerdu = false;          // Variable qui mémorisera le fait que le signal radio de l'émetteur n'est plus reçu ici, si c'est le cas

RF24 radio(pinCE_RF24, pinCSN_RF24);                // Instanciation du NRF24L01


//************************************************************//
// Format des données à recevoir, par ondes radio (structure) //
//************************************************************//
struct DonneesArecevoir {
  int valAvancerReculer;    // int (2 octets), transportant une valeur comprise entre -255 et +255  (-255 signifie ici reculer à vitesse maximale, et +255, avancer à vitesse maximale)
  int valDroiteGauche;      // int (2 octets), transportant une valeur comprise entre -255 et +255  (-255 signifie ici tourner à gauche pleinement, et +255, signifie tourner pleinement à droite)
};
DonneesArecevoir donnees;


//*******//
// SETUP //
//*******//
void setup() {

  // Configuration des pins de l'Arduino Nano à destination du module L298N, en sorties
  pinMode(borneENA_L298N, OUTPUT);
  pinMode(borneIN1_L298N, OUTPUT);
  pinMode(borneIN2_L298N, OUTPUT);
  pinMode(borneIN3_L298N, OUTPUT);
  pinMode(borneIN4_L298N, OUTPUT);
  pinMode(borneENB_L298N, OUTPUT);
  
  // Configuration des pins de l'Arduino en destination des LEDs, en sorties également
  pinMode(pinLed_BatterieFaible, OUTPUT);       // Sortie LED rouge "Batterie faible"
  pinMode(pinLed_SignalRadioPerdu, OUTPUT);     // Sortie LED orange "Signal radio perdu"
  
  // Allumage/extinction des LEDS au démarrage, pour "autotest"
  autotestLedsAuDemarrage();
  arretMoteurs();
  
  // Démarrage du module NRF24L01+, avec paramétrages de base
  radio.begin();                        // Initialisation du module nRF24
  radio.openReadingPipe(0, adresse);    // Ouverture du tunnel en "LECTURE" (avec le "nom" qu'on lui a donné)
  radio.setPALevel(RF24_PA_MIN);        // Sélection d'un niveau "MINIMAL" d'émission, pour communiquer (car pas besoin ici d'une forte puissance, pour ce projet)
  radio.startListening();               // Activation du mode "écoute" pour le nRF24 (signifiant qu'on va recevoir des données, et non en émettre, ici)

}

//**************************//
// Boucle principale : LOOP //
//**************************//
void loop() {

  mesurerTensionsAccus();    // Stocke le résultat dans les variables : tensionAccu1, tensionAccu2, et tensionAccu3

  // Test : vérification de la tension de chaque accu, par rapport à la valeur minimale qu'on a définit plus haut
  if(tensionAccu1 < tensionMinimaleAccu || tensionAccu2 < tensionMinimaleAccu || tensionAccu3 < tensionMinimaleAccu) {
    // Si un des accus a une tension trop basse, alors "on arrête tout"
    stopperProgramme();
  }
  else {
    // Si les tensions d'accus sont bonnes, alors on continue le programme
    if(donneesRecues()) {
      heureDernierSignalRecu = millis();                  // Mise à jour de "l'heure" du dernier signal reçu
      digitalWrite(pinLed_SignalRadioPerdu, HIGH);        // Extinction de la LED "signal radio perdu", si allumée
      //formatageDesDonneesRecues();
      envoiConsignesAuxMoteurs();
    } else {
      // Si aucune donnée radio n'est reçue, alors on regarde si ça fait plus de 2 secondes (soit 2000 millisecondes)
      if((millis() - heureDernierSignalRecu) > 2000) {
        // Cela fait plus de 2 seconde qu'aucun signal radio n'a été reçu
        digitalWrite(pinLed_SignalRadioPerdu, LOW);       // Allumage de la LED "signal radio perdu"
        arretMoteurs();                                   // Arrêt des moteurs, car aucune donnée reçue (émetteur coupé ? trop éloigné ?)
      }
    }
  }

}



//**************************************//
// Fonction : autotestLedsAuDemarrage() //
//**************************************//
void autotestLedsAuDemarrage() {
  
  // Allumage puis extinction des LEDs rouge et orange (activent à l'état bas), et ce, deux fois de suite
  for(byte i = 0 ; i < 3 ; i++) {
    digitalWrite(pinLed_BatterieFaible, LOW);
    digitalWrite(pinLed_SignalRadioPerdu, LOW);
    delay(20);
    digitalWrite(pinLed_BatterieFaible, HIGH);
    digitalWrite(pinLed_SignalRadioPerdu, HIGH);
    delay(200);    
  }
  
}

//***********************************//
// Fonction : mesurerTensionsAccus() //
//***********************************//
void mesurerTensionsAccus() {
  
  // Lecture des entrées analogiques
  valBrocheVA = analogRead(pinVA_PACK_ACCUS_18650);       // Valeur comprise entre 0 et 1023 (avec 1023 correspondant à +5V)
  valBrocheVB = analogRead(pinVB_PACK_ACCUS_18650);       // Valeur comprise entre 0 et 1023 (avec 1023 correspondant à +5V)
  valBrocheVC = analogRead(pinVC_PACK_ACCUS_18650);       // Valeur comprise entre 0 et 1023 (avec 1023 correspondant à +5V)

  // Calcul des tensions réelles, correspondantes à ces entrées (avec calculs inverses, dûs aux ponts diviseurs de tension)
  // Les valeurs obtenues seront, au nominal, de : 11,1V pour VA / 7,4V pour VB / 3,7V pour VC
  valTensionVA = (float)valBrocheVA * (valeurResistanceDiviseur_serie + valeurResistanceDiviseur_pullDown) / valeurResistanceDiviseur_pullDown * tensionReferenceAdcArduino / 1023;
  valTensionVB = (float)valBrocheVB * (valeurResistanceDiviseur_serie + valeurResistanceDiviseur_pullDown) / valeurResistanceDiviseur_pullDown * tensionReferenceAdcArduino / 1023;
  valTensionVC = (float)valBrocheVC * (valeurResistanceDiviseur_serie + valeurResistanceDiviseur_pullDown) / valeurResistanceDiviseur_pullDown * tensionReferenceAdcArduino / 1023;

  // Calcul des tensions individuelles de chaque accu lithium ion
  tensionAccu1 = valTensionVA - valTensionVB;             // 11,1 - 7,4 = 3,7V au nominal
  tensionAccu2 = valTensionVB - valTensionVC;             //  7,4 - 3,7 = 3,7V au nominal
  tensionAccu3 = valTensionVC;                            //  3,7 - 0,0 = 3,7V au nominal
  
}

//*******************************//
// Fonction : stopperProgramme() //
//*******************************//
void stopperProgramme() {
    digitalWrite(pinLed_BatterieFaible, LOW);         // On allume la LED rouge ("batterie faible")
    digitalWrite(pinLed_SignalRadioPerdu, HIGH);      // On éteint la LED orange ("signal radio perdu"), si allumée
    arretMoteurs();                                   // On met les moteurs au repos
    radio.powerDown();                                // Mise en sommeil du module récepteur nRF24
    while(1) {}                                       // Boucle sans fin ("arrêt du programme")
}

//****************************//
// Fonction : donneesRecues() //
//****************************//
boolean donneesRecues() {

  // On regarde si des données sont en attente de lecture
  if (radio.available()) {
    // Si oui, on les lit, et on retourne "VRAI"
    radio.read(&donnees, sizeof(donnees)); 
    return true;    
  } else {
    // Sinon, on retourne "FAUX"
    return false; 
  }

}

//***************************//
// Fonction : arretMoteurs() //
//***************************//
void arretMoteurs() {

  // Stoppe les signaux PWM de commande
  analogWrite(borneENA_L298N, 0);
  analogWrite(borneENB_L298N, 0);

  // Mise à l'état bas des autres sorties (non indispensable)
  digitalWrite(borneIN1_L298N, LOW);
  digitalWrite(borneIN2_L298N, LOW);
  digitalWrite(borneIN3_L298N, LOW);
  digitalWrite(borneIN4_L298N, LOW);
  
}

//****************************************//
// Fonction : formatageDesDonneesRecues() //
//****************************************//
void formatageDesDonneesRecues() {
  
}

//***************************************//
// Fonction : envoiConsignesAuxMoteurs() //
//***************************************//
void envoiConsignesAuxMoteurs() {
  
  // Analyse des données, et définition de valeurs qui vont nous servir
  byte vitesse = map(abs(donnees.valAvancerReculer), 0, 255, vitesseMinimale, vitesseMaximale);
  boolean bMarcheAvant = donnees.valAvancerReculer > 0 ? true : false;
  boolean bMarcheArriere = donnees.valAvancerReculer < 0 ? true : false;
  boolean bAuPointMort = donnees.valAvancerReculer == 0 ? true : false;

  float coefficienAngleVolant = (1 - abs(donnees.valDroiteGauche) / 255 * 0.7);
  boolean bTournerAdroite = donnees.valDroiteGauche > 0 ? true : false;
  boolean bTournerAgauche = donnees.valDroiteGauche < 0 ? true : false;


  // Cas "MARCHE AVANT"
  if (bMarcheAvant) {
    // Indication de sens des ponts en H, des L298N
    digitalWrite(borneIN1_L298N, HIGH);       // L'entrée IN1 doit être au niveau haut
    digitalWrite(borneIN2_L298N, LOW);        // L'entrée IN2 doit être au niveau bas
    digitalWrite(borneIN3_L298N, HIGH);       // L'entrée IN3 doit être au niveau haut
    digitalWrite(borneIN4_L298N, LOW);        // L'entrée IN4 doit être au niveau bas

    // Vitesse du moteur de droite (suivant consigne, sauf si besoin de tourner à droite ; là, il faudra le ralentir)
    if(bTournerAdroite) {
      analogWrite(borneENA_L298N, vitesse * coefficienAngleVolant);
    } else {
      analogWrite(borneENA_L298N, vitesse);
    }

    // Vitesse du moteur de gauche (suivant consigne, sauf si besoin de tourner à gauche ; là, il faudra le ralentir)
    if(bTournerAgauche) {
      analogWrite(borneENB_L298N, vitesse * coefficienAngleVolant);
    } else {
      analogWrite(borneENB_L298N, vitesse);
    }
  
  }
  
  // Cas "MARCHE ARRIERE"
  if (bMarcheArriere) {
    // Indication de sens des ponts en H, des L298N
    digitalWrite(borneIN1_L298N, LOW);        // L'entrée IN1 doit être au niveau bas
    digitalWrite(borneIN2_L298N, HIGH);       // L'entrée IN2 doit être au niveau haut
    digitalWrite(borneIN3_L298N, LOW);        // L'entrée IN3 doit être au niveau bas
    digitalWrite(borneIN4_L298N, HIGH);       // L'entrée IN4 doit être au niveau haut

      // Vitesse du moteur de droite (suivant consigne, sauf si besoin de tourner à droite ; là, il faudra le ralentir)
    if(bTournerAdroite) {
      analogWrite(borneENA_L298N, vitesse * coefficienAngleVolant);
    } else {
      analogWrite(borneENA_L298N, vitesse);
    }

    // Vitesse du moteur de gauche (suivant consigne, sauf si besoin de tourner à gauche ; là, il faudra le ralentir)
    if(bTournerAgauche) {
      analogWrite(borneENB_L298N, vitesse * coefficienAngleVolant);
    } else {
      analogWrite(borneENB_L298N, vitesse);
    }

  }

  // Cas "AU POINT MORT"
  if (bAuPointMort) {
    arretMoteurs();
  }

}
